/* AMX Mod script.
*
* (c) Copyright 2002, OLO
* This file is provided as is (no warranties).
*
*/

#include <amxmod>

/* cvars:
*  sets how many slots we want to keep for admins
*  amx_reserved_slots < value >
*
*  sets kicking rules to keep one free place on server
*  unless amx_reserved_slots is reached
*  amx_reservation < value >
*  0 - don't kick anyone and keep reserved
*      slots only for players with reservation
*  1 - kick by shortest playing time
*  2 - kick by worst ping
*  3 - don't kick anyone and put on remaining
*      reserved slots only players with reservation
*/

new lastcon[32]
new lastcon_num
new Float:lastcon_time

new logfilename[256]

kick_player(id){
	// gather some info
	new name[32], authid[16]
	get_user_name(id,name,31)
	get_user_authid(id,authid,15)
	// save who and how often was connecting
	new Float:gametime = get_gametime()
	if (!equal(lastcon,name) || lastcon_time + 5.0 < gametime )	{
		copy(lastcon,31,name)
		lastcon_num = 0
	}
	else
		++lastcon_num
	lastcon_time = gametime
	new userid = get_user_userid(id)
	if (is_user_bot(id)) {// if bot then kick with console cmd
		server_cmd("kick #%d",userid)
		log_to_file(logfilename,"Reservation: ^"%s<%d><%s><>^" drop due to slot reservation",
			name,userid,authid)
	}
	else if ( lastcon_num > 3 ) {// reconnecter - ban him
		server_cmd("banid 1 #%d",userid)
		if (lastcon_num < 5) {
			client_cmd(id,"echo * You have been banned for 1 minute")
			log_to_file(logfilename,"Reservation: ^"%s<%d><%s><>^" ban reconnecter for 1 minute",
				name,userid,authid)
		}
		client_cmd(id,"disconnect")
	}
	else {// else disconnect with leaving the proper reason on client console
		if ( lastcon_num > 2 )
			client_cmd(id,"echo * Next time you will be banned for 1 minute")
		client_cmd(id,"echo * Dropped due to slot reservation;disconnect")
		log_to_file(logfilename,"Reservation: ^"%s<%d><%s><>^" drop due to slot reservation",
			name,userid,authid)
	}
}

kick_lag(){
	new who = 0, ping, loss, worst = 0
	new maxplayers = get_maxplayers()
	for(new i = 1; i <= maxplayers; ++i){
		if ( !is_user_connected(i) && !is_user_connecting(i) ) continue // not used slot	
		if (get_user_flags(i)&ADMIN_RESERVATION) continue // has reservation (don't touch)
		get_user_ping(i,ping,loss) // get ping
		if ( ping > worst ) {
			worst = ping
			who = i
		}
	}
	if (who) kick_player(who)
}

kick_fresh(){
	new who = 0, time, shortest = 999999999
	new maxplayers = get_maxplayers()
	for(new i = 1; i <= maxplayers; ++i){
		if ( !is_user_connected(i) && !is_user_connecting(i) )	continue // not used slot	
		if (get_user_flags(i)&ADMIN_RESERVATION) continue // has reservation (don't touch)
		time = get_user_time(i) // get user playing time with connection duration	
		if ( shortest > time )	{
			shortest = time
			who = i
		}
	}
	if (who) kick_player(who)
}

get_slots_usage(&pnum,&rnum) {
	new maxplayers = get_maxplayers()
	pnum = rnum = 0
	for(new i = 1; i <= maxplayers; ++i)	{
		if ( !is_user_connected(i) && !is_user_connecting(i) ) continue // not used slot
		++pnum
		if (get_user_flags(i)&ADMIN_RESERVATION)	++rnum
	}
}

/* If there is at least one place then allow enter the game anyone.
* If the slots are reserved and not all filled then kick
* with amx_reservation rules.
*/
public client_connect(id){
	new rmax, rnum
	switch(check_slots(id,rnum,rmax))	{
		case 1:{
			client_cmd(id,"echo ^"* You are using reserved slot (%d/%d)^"",rnum,rmax)
			new name[32], authid[16]
			get_user_name(id,name,31)
			get_user_authid(id,authid,15)
			log_to_file(logfilename,"Reservation: ^"%s<%d><%s><>^" is using reserved slot (slot ^"%d^") (max ^"%d^")",
				name,get_user_userid(id),authid,rnum,rmax)
		}
		case 2: // this player have to be kicked - no free slots
			return PLUGIN_HANDLED // block connect in other plugins (especially in consgreet)
	}

	update_real_reserved_slots(rnum,rmax)

	return PLUGIN_CONTINUE
}

public client_disconnect(id){
	new rmax = get_cvar_num("amx_reserved_slots")
	new nplayers, rnum
	get_slots_usage(nplayers,rnum)
	if (get_user_flags(id)&ADMIN_RESERVATION)
		rnum--
	update_real_reserved_slots(rnum,rmax)
	return PLUGIN_CONTINUE
}

check_slots(id,&rnum,&rmax){
	new nplayers, restype = get_cvar_num("amx_reservation")
	new maxplayers = get_maxplayers()
	rmax = get_cvar_num("amx_reserved_slots")
	get_slots_usage(nplayers,rnum)
	// reserved - reserved_used ( 2 - 1 (already one player with res. on server) = 1 res. slot remaining )
	new resleft = rmax - rnum
	if (restype && restype < 3) {// kick if one slot left and player don't have reservation
		if (nplayers < maxplayers) // less players than maxplayers is set (i.e.: 8/12 or 2 /12)
			return 0 // return with no comments
		// there is 11/12 situation
		if (get_user_flags(id)&ADMIN_RESERVATION) {// has reservation
			if (resleft > 0) {// and still has we haven't reach reserved number (still res. slots remaining)
				switch(restype) {// depends on kick method free one slot
					case 1: kick_fresh()
					case 2: kick_lag()
				}
			}
			return 1 // return with signal that we are using reserved slot
		}
		if ( resleft > 0 ) {// res. slot remaining so kick us
			kick_player(id)	
			return 2 // return with signal to not send consgreet or other activity on connection
		}
	}
	// else don't kick - keep free slots
	else if (nplayers > maxplayers - rmax){
		if (get_user_flags(id)&ADMIN_RESERVATION)
			return 1 // if has reservation then return with signal about using res. slot
		if ( restype ? true : (resleft > 0) ) {// if res. slots still remain
			kick_player(id)
			return 2  // return with signal to not send consgreet or other activity on connection
		}
	}
	return 0
}

update_real_reserved_slots(rnum,rmax) {
	new real_res_slots = -1
	new restype = get_cvar_num("amx_reservation")
	new resleft = rmax - rnum
	if (restype && restype < 3) {
			real_res_slots = ( resleft > 0 ) ? 1 : 0
	} 
	else {
			real_res_slots = ( resleft < 0) ? 0 : resleft
	}
	set_cvar_num("amx_real_res_slots", real_res_slots)
}

public plugin_init(){
	register_plugin("Slots Reservation","0.8.7","default")
	register_cvar("amx_reserved_slots","2")
	register_cvar("amx_real_res_slots","-1",FCVAR_SERVER|FCVAR_EXTDLL|FCVAR_UNLOGGED|FCVAR_SPONLY)
	register_cvar("amx_reservation","1")
	get_time("addons/amx/logs/admin%m%d.log",logfilename,255)
	return PLUGIN_CONTINUE
}
